package ca.yorku.cse.mack.h4touch;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.res.Configuration;
import android.media.MediaPlayer;
import android.media.MediaScannerConnection;
import android.os.Bundle;
import android.os.Environment;
import android.os.Vibrator;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.EditText;
import android.widget.TextView;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Random;

import ca.yorku.cse.mack.h4touch.H4Keyboard.H4Listener;

/**
 * <h1>H4Touch</h1>
 *
 * <h3>Summary</h3>
 *
 * <ul>
 * <li>Android experiment software for the H4Writer text entry method using touch.
 * </ul>
 * <p>
 *
 * <h3>Related References</h3>
 *
 * The following publication presents research where a variation of this software was used.
 * <p>
 *
 * <ul>
 * <li>
 * <a href="http://www.yorku.ca/mack/ijvwhci2013.html">Eye
 * on the message: Reducing attention demand for touch-based text entry</a>, by MacKenzie and Castellucci (<i>Int
 * J Virtual Worlds and HCI 2013</i>).
 * <p>
 * </ul>
 * <p>
 *
 * The following publication describes the design of H4Writer, the text entry method upon
 * which
 * H4Touch is
 * based.
 * <p>
 *
 * <ul>
 * <li><a href="http://www.yorku.ca/mack/uist2011.html"> 1 thumb, 4 buttons, 20 words per minute:
 * Design and evaluation of H4-Writer</a>, by MacKenzie, Soukoreff, and Helga (<i>UIST 2011</i>).
 * <p>
 * </ul>
 * <p>
 *
 * <h3>Setup Parameters</h3>
 *
 * The application launches with a setup dialog: (click to enlarge)<p>
 *
 * <center> <a href="H4Touch-1.jpg"><img src="H4Touch-1.jpg" width="200" alt="image"></a> </center> <p>
 *
 * In all, there are 13 setup parameters, organized in eight spinners and five checkboxes. The setup parameters are
 * explained as follows. <p>
 *
 * <blockquote>
 * <table border="1" cellspacing="0" cellpadding="6" valign="top">
 *
 * <tr bgcolor="#cccccc">
 * <th align="center">Parameter
 * <th align="center">Description
 *
 * <tr valign="top"> <td>Participant Code <td>A code to identify the user being tested. In experimental research, a user
 * is typically called a "participant". <p>
 *
 * <tr valign="top"> <td>Session Code <td>A code to identify the session. This code could be used in an a longitudinal
 * experiment, where participants are tested over multiple sessions. Each session might involve an hour of testing, with
 * one session per day over multiple days. <p>
 *
 * <tr valign="top">
 * <td>Block Code
 * <td>Note: There is an inactive setup parameter for "Block code". The block code is generated automatically by the
 * software. <p>
 *
 * <tr valign="top"> <td>Group Code <td>A code to identify the group. If the experiment uses a within-subjects design,
 * participants are usually assigned to groups. This code identifies the group. This procedure is known as
 * <i>counterbalancing</i>, which is used to offset order effects. (See section 5.11 in the course's Suggested Readings
 * for complete details.) <p>
 *
 * <tr valign="top"> <td>Condition Code <td>A code to identify an external condition. This setup item is an editable
 * text field. Tap to popup an input method editor to change the value. <p>
 *
 * The condition code is useful if the experiment included an independent variable pertaining to an environmental
 * condition or a context of use. As an example, an experiment might include "mobility" as an independent variable with
 * levels "sitting" and "walking". The code "Sitting" might be used for the sitting condition and the code "Walking" for
 * the walking condition. These codes appear in the output data to facilitate analyses. BTW, an example of an experiment
 * using this independent variable is described by Fitton et al. in <a href="http://www.yorku.ca/mack/bhci2013
 * .html">Exploring Tilt-Based Text Input For Mobile Devices With Teenagers</a>. <p>
 *
 * <tr valign="top"> <td>Phrases per Block<td>The number of phrases entered for each block of
 * testing.
 * <p>
 *
 * <tr valign="top"> <td>Phrases File <td>A file containing phrases of text that are selected at random and presented to
 * the user for input. The following options appear: <p>
 *
 * <ul>
 * <li>phrases2 - a collection of 500 phrases (no punctuation)
 * <li>quickbrownfox - the phrase "the quick brown fox jumps over the lazy dog"
 * </ul> <p>
 *
 * Note: The actual phrases are in files in <code>/res/values/</code>. <p>
 *
 * <tr valign="top">
 * <td>Key style
 * <td>Sets the arrangement of the 4 keys.  The options are "+" or "x", corresponding, in order, to<p>
 *
 * <center>
 * <a href="H4Touch-2.jpg"><img src="H4Touch-2.jpg" width=100 alt="image"></a>
 * <a href="H4Touch-3.jpg"><img src="H4Touch-3.jpg" width=100 alt="image"></a>
 * </center>
 *
 * <tr valign="top">
 * <td>Vibrotactile feedback
 * <td>A checkbox item. If selected, enables a vibrotactile pulse to be emitted on each key selection.
 *
 * <tr valign="top">
 * <td>Auditory feedback
 * <td>A checkbox item. If selected, enables an auditory click to be emitted on each key
 * selection.
 *
 * <tr valign="top">
 * <td>Show presented text during input
 * <td>A checkbox item. If selected, shows the presented text during entry.  If un-selected, the presented text phrase
 * disappears as soon as input begins for the current text phrase.
 *
 * <tr valign="top">
 * <td>Lowercase only
 * <td>A checkbox item. If selected, converts to lowercase any uppercase letters in the presented text phrase.
 *
 * <tr valign="top">
 * <td>Show Huffman codes
 * <td>A checkbox item. If selected, letters are shown on keys as entry progresses.  At the beginning of each character
 * of input, the letters appear as follows (as per the Key Style parameter): (click to enlarge)
 * <p>
 *
 * <center>
 * <a href="H4Touch-4.jpg"><img src="H4Touch-4.jpg" width=100 alt="image"></a>
 * <a href="H4Touch-5.jpg"><img src="H4Touch-5.jpg" width=100 alt="image"></a>
 * </center>
 *
 * The letters are updated on each key
 * press to show the progression through the Huffman code tree.  If un-selected, letters do not appear on
 * keys as entry progresses.  In this case, the keys appear as seen above for the Key Style parameter.
 * </table> </blockquote> <p>
 *
 * <h3>Program Operation</h3>
 *
 * After the setup parameters are chosen, the program is launched by clicking the OK button.  Assuming the default
 * setup, the following screen appears (below left):
 * <p>
 *
 * <center>
 * <a href="H4Touch-6.jpg"><img src="H4Touch-6.jpg" width=150 alt="image"></a>
 * <a href="H4Touch-7.jpg"><img src="H4Touch-7.jpg" width=150 alt="image"></a>
 * <a href="H4Touch-8.jpg"><img src="H4Touch-8.jpg" width=150 alt="image"></a>
 * </center>
 * <p>
 *
 * A phrase of text is drawn at random from the phrases file and presented for input.  To enter text, the user taps
 * keys according the desired letter.  For the first letter in the phrase above ("h"), the user first taps the
 * top-left key. The key labels are updated to show the progression through the Huffman tree (above center).  The
 * desired letter now appears on the bottom-left key.  The user taps the key whereupon the letter appears in
 * the transcribed text field (above right).  For this example, entering the letter "h" required two taps.
 * Depending on the
 * letter, sometimes more taps are required.  On average, 2.321 taps are required for each letter entered.
 * <p>
 *
 * At the end of a phrase, results appear:
 * <p>
 *
 * <center>
 * <a href="H4Touch-9.jpg"><img src="H4Touch-9.jpg" width=200 alt="image"></a>
 * </center>
 * <p>
 *
 * <h3>Data Storage</h3>
 *
 * Data are also saved in files for follow-up analyses.  There are two files, an "sd1" file which contains data on a
 * per-keystroke basis, and an "sd2" file which contains summary data for each phrase of input.  The data files are
 * saved in
 * the public directory on Android devices in a sub-directory called "H4TouchData".  Below are example
 * files:
 * <p>
 *
 * <ul>
 * <li>sd1 &ndash; <a href="H4Touch-P01-S01-B02-G01-C01-SQ.sd1">H4Touch-P01-S01-B02-G01-C01-SQ.sd1</a>
 * <li>sd2 &ndash; <a href="H4Touch-P01-S01-B02-G01-C01-SQ.sd2">H4Touch-P01-S01-B02-G01-C01-SQ.sd2</a>
 * </ul>
 * <p>
 *
 * The data in the sd2 files are comma-delimited to facilitate importing in Excel or a statistics
 * application for follow-up analyses.  An example is shown below for the data in the sd2 file above.
 * <p>
 *
 * <center>
 * <a href="H4Touch-10.jpg"><img src="H4Touch-10.jpg" width=800 alt="image"></a>
 * </center>
 *
 * @author Scott MacKenzie, 2014-2017
 */

public class H4TouchActivity extends Activity implements H4Listener
{
	final String MYDEBUG = "MYDEBUG";
	final String APP = "H4Touch";
	final String WORKING_DIRECTORY = "/H4TouchData/";
	final String SD2_HEADER = "App,Participant,Session,Block,Group,Condition,Key_style,"
			+ "Min_keystrokes,Actual_keystrokes,Presented_char,Transcribed_char,Time(s),MSD,"
			+ "Entry_speed(wpm),Error_rate(%),KSPC,Efficiency\n";

	final int KEYSTROKES_FOR_ENTER = 2; // adjustment needed for output data (keystrokes for ENTER not counted)
	final long PULSE_DURATION = 5;
	final int MARGIN_LEFT = 5; // should be consistent with values in main.xml

	// setup parameters
	String participantCode, sessionCode, blockCode, groupCode, conditionCode;
	String[] phrases;
	boolean vibrotactileFeedback, auditoryFeedback;
	boolean showText;
	boolean lowercaseOnly, showHuffman;
	// end setup parameters

	boolean phraseStarted;
	boolean newPhrase = true;
	boolean done = false;
	Vibrator vib;
	MediaPlayer enter, tick, miss;
	BufferedWriter sd1, sd2;
	File f1, f2;
	String sd2Leader;
	EditText presentedTextField, transcribedTextField;
	H4Keyboard h4k;
	String presentedText, transcribedText;
	Random r = new Random();
	int keystrokeCount, phraseCount, phrasesPerBlock;
	long timeStampStart;
	int timeForPhrase;
	int keyStyle;
	String keyStyleString;
	ArrayList<String> samples;

	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);

		Bundle b = getIntent().getExtras();
		participantCode = b.getString("participantCode");
		sessionCode = b.getString("sessionCode");
		blockCode = b.getString("blockCode");
		groupCode = b.getString("groupCode");
		conditionCode = b.getString("conditionCode");
		phrasesPerBlock = b.getInt("phrasesPerBlock");
		vibrotactileFeedback = b.getBoolean("vibrotactileFeedback");
		auditoryFeedback = b.getBoolean("auditoryFeedback");
		showText = b.getBoolean("showText");
		lowercaseOnly = b.getBoolean("lowercaseOnly");
		showHuffman = b.getBoolean("showHuffmanCodes");

		// special fiddling required for phrases array
		if (b.getString("phrasesFile").equals("phrases2"))
			phrases = getResources().getStringArray(R.array.phrases2);
		else if (b.getString("phrasesFile").equals("quickbrownfox"))
			phrases = getResources().getStringArray(R.array.quickbrownfox);
		else if (b.getString("phrasesFile").equals("phrases100"))
			phrases = getResources().getStringArray(R.array.phrases100);
		else if (b.getString("phrasesFile").equals("alphabet"))
			phrases = getResources().getStringArray(R.array.alphabet);

		// special fiddling required for key style
		keyStyle = b.getString("keyStyle").equals("x") ? H4Keyboard.TRIANGLE : H4Keyboard.SQUARE;
		if (keyStyle == H4Keyboard.TRIANGLE)
			keyStyleString = "TR";
		else
			keyStyleString = "SQ";

		// ===================
		// File initialization
		// ===================

		// make a working directory (if necessary) to store data files
		File dataDirectory = new File(Environment.getExternalStorageDirectory() + WORKING_DIRECTORY);
		if (!dataDirectory.exists() && !dataDirectory.mkdirs())
		{
			Log.e(MYDEBUG, "ERROR --> FAILED TO CREATE DIRECTORY: " + WORKING_DIRECTORY);
			super.onDestroy(); // cleanup
			this.finish(); // terminate
		}
		Log.i(MYDEBUG, "Working directory=" + dataDirectory);

		/**
		 * The following do-loop creates output data files and a string leader to write to the sd2
		 * output data file.  Both the filenames and the leader are constructed by combining the setup parameters
		 * so that the filenames and leader are unique and also reveal the conditions used for the block of input.
		 *
		 * The block code begins "B01" and is incremented on each loop iteration until an available
		 * filename is found.  The goal, of course, is to ensure data files are not inadvertently overwritten.
		 */
		int blockNumber = 0;
		do
		{
			++blockNumber;
			String blockCode = String.format(Locale.CANADA, "B%02d", blockNumber);
			String baseFilename = String.format("%s-%s-%s-%s-%s-%s-%s", APP, participantCode,
					sessionCode, blockCode, groupCode, conditionCode, keyStyleString);
			f1 = new File(dataDirectory, baseFilename + ".sd1");
			f2 = new File(dataDirectory, baseFilename + ".sd2");

			// also make a comma-delimited leader that will begin each data line written to the sd2 file
			sd2Leader = String.format("%s,%s,%s,%s,%s,%s,%s", APP, participantCode, sessionCode,
					blockCode, groupCode, conditionCode, keyStyleString);
		} while (f1.exists() || f2.exists());

		try
		{
			sd1 = new BufferedWriter(new FileWriter(f1));
			sd2 = new BufferedWriter(new FileWriter(f2));

			// output header in sd2 file
			sd2.write(SD2_HEADER, 0, SD2_HEADER.length());
			sd2.flush();
		} catch (IOException e)
		{
			super.onDestroy();
			this.finish();
		}
		// end file initialization

		presentedTextField = (EditText)findViewById(R.id.presented);
		transcribedTextField = (EditText)findViewById(R.id.transcribed);

		// prevent system's IME soft keyboard from popping up
		getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);

		h4k = (H4Keyboard)findViewById(R.id.h4keys);
		h4k.setH4Listener(this);
		h4k.showHuffmanCodes = showHuffman;
		h4k.keyStyle = keyStyle;

		// determine display width and height
		DisplayMetrics dm = new DisplayMetrics();
		getWindowManager().getDefaultDisplay().getMetrics(dm);
		float scalingFactor = dm.density;
		int width = dm.widthPixels - (int) (2 * MARGIN_LEFT * scalingFactor + 0.5f);
		h4k.setKeyboardSize(width, dm.heightPixels, scalingFactor);

		// initialize vibrotactile and auditory feedback objects
		vib = (Vibrator)getSystemService(Context.VIBRATOR_SERVICE);
		tick = MediaPlayer.create(getApplicationContext(), R.raw.click1);
		miss = MediaPlayer.create(getApplicationContext(), R.raw.miss);
		enter = MediaPlayer.create(getApplicationContext(), R.raw.blip1);

		samples = new ArrayList<String>();
		phraseCount = 0;
		prepareForPhrase();
	}

	public void H4TapEvent(H4Keyboard.H4Event h4e)
	{
		long timeStamp = h4e.timeStamp;
		String symbol = h4e.symbol;

		if (!phraseStarted)
		{
			phraseStarted = true;
			timeStampStart = h4e.timeStamp;
		}

		// save info on this H4Touch event (will be written to sd1 file)
		samples.add(String.format(Locale.CANADA, "%d %d %s", h4e.n, (h4e.timeStamp - timeStampStart), h4e.symbol));

		++keystrokeCount;

		if (done)
			doDone();

		if (symbol.equals(""))
			return; // nothing to process yet

		if (!symbol.equals(H4Keyboard.ENTER))
		{
			// first, update the transcribed text field, as appropriate, and output the appropriate feedback
			if (symbol.equals(H4Keyboard.SPACE))
			{
				transcribedText += " ";
				if (vibrotactileFeedback)
					vib.vibrate(PULSE_DURATION);
				if (auditoryFeedback)
					tick.start();
			} else if (symbol.equals(H4Keyboard.BACKSPACE))
			{
				if (transcribedText.length() >= 1)
				{
					transcribedText = transcribedText.substring(0, transcribedText.length() - 1);
					if (vibrotactileFeedback)
						vib.vibrate(PULSE_DURATION);
					if (auditoryFeedback)
						tick.start();
				} else if (auditoryFeedback)
					miss.start();
			} else
			{
				transcribedText += symbol;
				if (vibrotactileFeedback)
					vib.vibrate(PULSE_DURATION);
				if (auditoryFeedback)
					tick.start();
			}
			transcribedTextField.setText(transcribedText);
			transcribedTextField.setSelection(transcribedText.length()); // move cursor to end

		} else
		// ENTER (end of phrase)
		{
			doEndOfPhrase(timeStamp);
		}
	}

	private void doEndOfPhrase(long timeStampEnd)
	{
		if (vibrotactileFeedback)
			vib.vibrate(PULSE_DURATION);
		if (auditoryFeedback)
			enter.start();

		newPhrase = true;
		StringBuilder resultsString = new StringBuilder(500);
		resultsString.append("Thank you!\n");

		// don't count the keystrokes for ENTER
		keystrokeCount -= KEYSTROKES_FOR_ENTER;
		timeForPhrase = (int) (timeStampEnd - timeStampStart);

		String s1 = presentedText;
		String s2 = transcribedText.trim();
		if (lowercaseOnly)
		{
			s1 = s1.toLowerCase(Locale.US);
			s2 = s2.toLowerCase(Locale.US);
		}

		resultsString.append(String.format("Presented...\n   %s\n", s1));
		resultsString.append(String.format("Transcribed...\n   %s\n", s2));

		StringBuilder sd2Stuff = new StringBuilder(500);
		sd2Stuff.append(String.format("%s,", sd2Leader)); // begin with the leader created earlier

		// MINIMUM NUMBER OF KEYSTROKES
		sd2Stuff.append(String.format(Locale.CANADA, "%d,", minimumKeystrokes(presentedText)));

		// ACTUAL NUMBER OF KEYSTROKES
		sd2Stuff.append(String.format(Locale.CANADA, "%d,", keystrokeCount));

		// PRESENTED CHARACTERS (PHRASE LENGTH)
		sd2Stuff.append(String.format(Locale.CANADA, "%d,", s1.length()));

		// NUMBER OF CHARACTERS OF TEXT PRODUCED
		sd2Stuff.append(String.format(Locale.CANADA, "%d,", s2.length()));

		// TIME (SECONDS)
		float d = timeForPhrase / 1000.0f;
		sd2Stuff.append(String.format(Locale.CANADA, "%.2f,", d));

		// MSD
		MSD s1s2 = new MSD(s1.toLowerCase(Locale.US).trim(), s2.toLowerCase(Locale.US).trim());
		int msd = s1s2.getMSD();
		sd2Stuff.append(String.format(Locale.CANADA, "%d,", msd));

		// ENTRY SPEED (WPM)
		d = wpm(s2, timeForPhrase);
		sd2Stuff.append(String.format(Locale.CANADA, "%.2f,", d));
		resultsString.append(String.format(Locale.CANADA, "Entry speed = %.2f wpm\n", d));

		// ERROR RATE (%)
		d = (float)s1s2.getErrorRateNew();
		sd2Stuff.append(String.format(Locale.CANADA, "%.2f,", d));
		resultsString.append(String.format(Locale.CANADA, "Error rate = %.2f%%\n", d));

		// KSPC
		float kspc = (float)keystrokeCount / s2.length();
		sd2Stuff.append(String.format(Locale.CANADA, "%.4f,", kspc));

		// EFFICIENCY
		float kspcmin = (float)minimumKeystrokes(presentedText) / s1.length();
		float efficiency = (kspcmin / kspc) * 100.0f;
		resultsString.append(String.format(Locale.CANADA, "Extra taps = %d",(keystrokeCount - minimumKeystrokes(presentedText))));
		sd2Stuff.append(String.format(Locale.CANADA, "%.2f\n", efficiency));

		// present results to user
		showResultsDialog(resultsString.toString());

		// build sd1 data
		StringBuilder sd1Stuff = new StringBuilder(500);
		sd1Stuff.append(String.format("%s\n%s\n", presentedText, s2));

		for (String s : samples)
			sd1Stuff.append(String.format("%s\n", s));

		// dump data to sd1 and sd2 files
		try
		{
			sd1.write(sd1Stuff.toString(), 0, sd1Stuff.length());
			sd1.flush();
			sd2.write(sd2Stuff.toString(), 0, sd2Stuff.length());
			sd2.flush();
		} catch (IOException e)
		{
			// Log.d("MYDEBUG", "ERROR WRITING TO DATA FILE! e = " + e);
			this.finish();
		}

		// terminate if this was the last phrase in the block
		if (++phraseCount == phrasesPerBlock)
			done = true; // app will terminate on next touch (display results for last phrase)
		else
			prepareForPhrase();
	}

	private void showResultsDialog(String text)
	{
		LayoutInflater inflater = (LayoutInflater)this.getSystemService(LAYOUT_INFLATER_SERVICE);
		View layout = inflater.inflate(R.layout.results_dialog, (ViewGroup)findViewById(R.id.results_layout));

		// Set text
		TextView results = (TextView)layout.findViewById(R.id.resultsArea);
		results.setText(text);

		// Initialize the dialog
		AlertDialog.Builder parameters = new AlertDialog.Builder(this);
		parameters.setView(layout).setCancelable(false).setNeutralButton("OK", new DialogInterface.OnClickListener()
		{
			public void onClick(DialogInterface dialog, int id)
			{
				dialog.cancel(); // close this dialog
			}
		}).show();
	}

	private void doDone()
	{
		try
		{
			sd1.close();
			sd2.close();
			
			// Make the saved data files visible in Windows Explorer
			// There seems to be bug doing this with Android 4.4.  I'm using the following
			// code, instead of sendBroadcast.  See...
			// http://code.google.com/p/android/issues/detail?id=38282
			MediaScannerConnection.scanFile(this, new String[] { f1.getAbsolutePath(),
					f2.getAbsolutePath() }, null, null);
		} catch (IOException e)
		{
			// Log.i("MYDEBUG", "ERROR CLOSING DATA FILES! e = " + e);
		}
		super.onDestroy();
		this.finish();
	}

	/**
	 * Compute text entry speed in words per minute (wpm).
	 * 
	 * @param text
	 *            a phrase of text
	 * @param msTime
	 *            time to enter the phrase in milliseconds
	 * @return entry speed in words per minute (wpm) or -1 if time is <= 0
	 */
	public static float wpm(String text, long msTime)
	{
		float speed = text.length();
		if (msTime > 0)
			return speed / (msTime / 1000.0f) * 12.0f; // Note: 60 / 5 = 12
		else
			return -1f;
	}

	/**
	 * Compute the minimum number of keystrokes to enter a phrase.
	 * <p>
	 * 
	 * NOTE: The value returned excludes the keystrokes for a terminator, such as ENTER.
	 * <p>
	 * 
	 * @param phrase
	 *            the phrase (e.g., "the quick brown fox")
	 * @return minimum number of keystrokes to enter the phrase, or -1 if phrase length is zero
	 */
	private int minimumKeystrokes(String phrase)
	{
		int count = 0;
		for (int i = 0; i < phrase.length(); ++i)
		{
			Character c = phrase.charAt(i);
			String s = "" + c;

			// some pre-amble adjustments for SPACE, uppercase, and non-letters
			if (Character.isSpaceChar(c))
				s = H4Keyboard.SPACE; // re-map, if SPACE

			if (Character.isUpperCase(c)) // add transition to caps
			{
				count += sizeOfCode(H4Keyboard.SHIFT, H4Keyboard.LETTER_CODES);
				s = s.toLowerCase(Locale.US);
			}

			if (Character.isLetter(c) || Character.isSpaceChar(c) || s.equals(".") || s.equals(","))
			{
				count += sizeOfCode(s, H4Keyboard.LETTER_CODES);
			} else
			{
				count += sizeOfCode(H4Keyboard.SYMBOL, H4Keyboard.LETTER_CODES);
				count += sizeOfCode(s, H4Keyboard.SYMBOL_CODES);
			}
		}
		return count > 0 ? count : -1;
	}

	// compute the size of a huffman code for a given symbol.
	private int sizeOfCode(String symbol, String[][] huffmanCodes)
	{
		int count = 0;
		for (String[] s : huffmanCodes)
			if (symbol.equals(s[0]))
			{
				count += s[1].length();
				break;
			}

		//for (int i = 0; i < huffmanCodes.length; ++i)
		//{
		//	if (symbol.equals(huffmanCodes[i][0]))
		//	{
		//		count += huffmanCodes[i][1].length();
		//		break;
		//	}
		//}
		return count > 0 ? count : -1;
	}

	void prepareForPhrase()
	{
		String phrase = phrases[r.nextInt(phrases.length)];
		presentedText = lowercaseOnly ? phrase.toLowerCase(Locale.US) : phrase;
		presentedTextField.setText(presentedText);
		transcribedText = "";
		transcribedTextField.setText(transcribedText);
		phraseStarted = false;
		keystrokeCount = 0;
		samples.clear();
	}

	public void onConfigurationChanged(Configuration newConfig)
	{
		super.onConfigurationChanged(newConfig);

		// determine display width and height
		DisplayMetrics dm = new DisplayMetrics();
		getWindowManager().getDefaultDisplay().getMetrics(dm);
		float scalingFactor = dm.density;
		int width = dm.widthPixels - (int) (2 * MARGIN_LEFT * scalingFactor + 0.5f);
		h4k.setKeyboardSize(width, dm.heightPixels, scalingFactor);
	}

	public void onSaveInstanceState(Bundle savedInstanceState)
	{
		super.onSaveInstanceState(savedInstanceState);
	}

	public void onRestoreInstanceState(Bundle savedInstanceState)
	{
		super.onRestoreInstanceState(savedInstanceState);
	}
}